import datetime
import random
import textwrap
import warnings

import langchain_core
import langsmith.evaluation
import langsmith.schemas
import numpy as np
import prefect
import pydantic
import tqdm.auto

from ai_app.ai_utils import to_json_llm_input
from ai_app.apps.jira_service_desk.backend import Bot, get_system_prompt
from ai_app.apps.jira_service_desk.data import (
    JiraRequest,
    JiraRequestPrimaryKey,
    ManualJiraRequestGuide,
    fetch_use_case_guides_for_requests,
    get_saved_request_type_primary_keys,
)
from ai_app.apps.jira_service_desk.preparation import generate_requests
from ai_app.config import get_ai_postgres_engine, get_chat_model, get_judge_model_names
from ai_app.core import Response
from ai_app.external.atlassian import build_issue_url
from ai_app.utils import wrap_with_xml_tag


class JiraServiceDeskBotEvaluationExample(pydantic.BaseModel):
    generated_user_query: str
    issue_url: str
    model_used_for_generation: str
    request_type_primary_key: JiraRequestPrimaryKey
    issue_key: str


def try_create_evaluation_example_from_latest_resolved_request(
    request_primary_key: JiraRequestPrimaryKey,
) -> JiraServiceDeskBotEvaluationExample | None:
    requests = list(generate_requests(**request_primary_key.model_dump(), limit_requests=1))
    if len(requests) != 1:
        warnings.warn(
            f"Failed to fetch a singular request for primary key {request_primary_key}, "
            f"instead fetched {len(requests)} requests"
        )
        return

    request = requests[0]
    issue_key = request["Issue key"]
    judge_model_name = random.choice(get_judge_model_names())
    model = get_chat_model(judge_model_name)
    system_prompt = textwrap.dedent("""
        We are evaluating a Jira service desk support bot designed to help users find the correct 
        request type and assist with creating and filling out Jira requests. To test the bot, we 
        need realistic synthetic user messages that could have been sent to the bot before a request
        was created.

        **Your task**:

        - Review the provided resolved Jira request in JSON format.
        - Generate a message that a user might have sent to the bot before submitting this request.
        - The message should be natural and reflect how real users communicate: it may contain 
            errors, typos, be brief or incomplete, and written in a hurry.
        - Assume the user knows they are chatting with a bot, so the message is likely direct and 
            to the point, not polite or formal.
        - The message should be in the same language as the request description or comments.
        
        **Produce a single, realistic user message that could have initiated the provided Jira 
        request.**
    """)
    response = model.invoke(
        [
            langchain_core.messages.SystemMessage(system_prompt),
            langchain_core.messages.HumanMessage(to_json_llm_input(request)),
        ]
    )
    example = JiraServiceDeskBotEvaluationExample(
        request_type_primary_key=request_primary_key,
        issue_key=issue_key,
        issue_url=build_issue_url(issue_key=issue_key),
        generated_user_query=response.content,
        model_used_for_generation=judge_model_name,
    )
    return example


def get_or_create_jira_service_desk_dataset(
    dataset_name: str | None = None,
) -> langsmith.schemas.Dataset:
    dataset_name = dataset_name or f"Jira service desk {datetime.datetime.now():%Y-%m-%d %H:%M:%S}"
    client = langsmith.Client()
    if client.has_dataset(dataset_name=dataset_name):
        dataset = client.read_dataset(dataset_name=dataset_name)
        return dataset

    keys = get_saved_request_type_primary_keys(
        get_ai_postgres_engine(),
        JiraRequest.get_sa_primary_key() == ManualJiraRequestGuide.get_sa_primary_key(),
        ManualJiraRequestGuide.do_recommend,
    )
    examples = []
    for key in tqdm.auto.tqdm(keys):
        example = try_create_evaluation_example_from_latest_resolved_request(key)
        if not example:
            continue

        example = langsmith.schemas.ExampleCreate(
            inputs=example.model_dump(),
            metadata=example.model_dump(include=["model_used_for_generation", "issue_url"]),
        )
        examples.append(example)

    dataset = client.create_dataset(dataset_name=dataset_name)
    client.create_examples(dataset_name=dataset_name, examples=examples)
    return dataset


async def ask_jira_bot(example: JiraServiceDeskBotEvaluationExample, model_name: str) -> dict:
    user_message = langchain_core.messages.HumanMessage(content=example.generated_user_query)
    bot = Bot()
    response: Response = await bot.respond(
        model=get_chat_model(model_name),
        messages=[user_message],
    )
    response = {"ai_message": response.content_for_langsmith}
    return response


class JiraServiceDeskBotEvaluation(pydantic.BaseModel):
    evaluation: str = pydantic.Field(
        description=textwrap.dedent("""
            Evaluation of the Jira service desk bot response according to the system prompt.
            All text information should be in English, concise, clear, dry, and to the point.
        """),
    )
    score: float = pydantic.Field(
        description="The score of the Jira service desk bot response, from 0 to 10.",
        ge=0,
        le=10,
    )
    insight: str = pydantic.Field(
        description=textwrap.dedent("""
            A short, actionable insight or conclusion in English following the bot evaluation, for 
            example:
            - If the bot failed to recommend correct request type, why, and how to fix it.
            - How the bot behavior or its context can be improved.
        """),
    )


def evaluate_example(
    inputs: dict, outputs: dict, reference_outputs: dict
) -> langsmith.evaluation.EvaluationResults | langsmith.evaluation.EvaluationResult:
    example = JiraServiceDeskBotEvaluationExample(**inputs)
    requests = generate_requests(
        **example.request_type_primary_key.model_dump(),
        jql_filter=f"issue = {example.issue_key}",
    )
    requests = list(requests)
    if len(requests) != 1:
        raise RuntimeError(f"Failed to fetch the Jira request {example.issue_key}.")

    request = requests[0]
    guides = fetch_use_case_guides_for_requests([example.request_type_primary_key])
    if len(guides) > 1:
        raise RuntimeError(
            f"Expected to fetch zero or one guide for request type "
            f"{example.request_type_primary_key}, but fetched {len(guides)} guides."
        )

    guide = None if not guides else guides[0]
    context = {
        "Jira request": request,
        "Use case guide for the Jira request type": guide,
        "Generated user query": example.generated_user_query,
        "Jira service desk bot response": outputs,
    }

    jira_service_desk_system_prompt = wrap_with_xml_tag(
        "jira_service_desk_system_prompt", get_system_prompt(), with_new_lines=True
    )
    system_prompt = textwrap.dedent("""
        ## Role
        You are an expert quality assurance specialist evaluating a Jira Service Desk bot.
        Your assessment will help improve the bot's performance and user experience.
        
        ## Context
        You will be provided with the following context in JSON format:
        1. A real resolved Jira request.
        2. A synthetic user message that was generated based on this request, with which the user
            may have come to the Jira service desk bot, before creating the request.
        3. The Jira service desk bot response to the synthetic user message and the request type
            that the bot suggested.
        4. The use case guide for the Jira request type that the bot relies on when deciding 
            which request type to suggest. If the use case guide is of low quality or doesn't cover
            the provided request, it may explain why the bot failed to recommend the correct
            request type.

        ## Task
        Your task is to evaluate the quality of the Jira service desk bot response based on the
        following criteria:

        ### Criteria
        1. High importance:
            - **Adherence to instructions:** The response adheres to the bot's system prompt:
            {jira_service_desk_system_prompt}
        2. Medium importance:
            - **Correctness:** The suggested request type is the same as the one in the
                provided request. If the bot suggested a different request type, inspect 
                the provided request use case that the bot sees and try to understand why
                the bot failed to recommend the correct request type.
        3. Low importance:
            - **Relevance:** The response is directly relevant to the user's query.
            - **Clarity:** The response is clear and easy to understand.
            - **Usefulness:** The response provides useful information to the user.
            - **Professionalism:** The response is professional and courteous.
        
        ## Output Format
        Your output should contain the following information, written in English:
        1. Your though process and reasoning of the bot evaluation, according to the above criteria.
        2. A single score representing the quality of the bot response on a scale of 0 to 10.
        3. A short, actionable insight or conclusion following the bot evaluation.
    """).format(jira_service_desk_system_prompt=jira_service_desk_system_prompt)
    judge_model_input = [
        langchain_core.messages.SystemMessage(system_prompt),
        langchain_core.messages.HumanMessage(to_json_llm_input(context)),
    ]
    results = []
    score_values = []
    for judge_model_name in get_judge_model_names():
        judge_model = get_chat_model(judge_model_name)
        judge_model = judge_model.with_structured_output(JiraServiceDeskBotEvaluation)
        evaluation = judge_model.invoke(judge_model_input)
        score_values.append(evaluation.score)
        score = langsmith.evaluation.EvaluationResult(
            key=f"Score by {judge_model_name}", score=evaluation.score
        )
        insight = langsmith.evaluation.EvaluationResult(
            key=f"Insight by {judge_model_name}", value=evaluation.insight
        )
        results.extend([score, insight])

    average_result = langsmith.evaluation.EvaluationResult(
        key="Average score", score=np.mean(score_values)
    )
    results.append(average_result)
    results = langsmith.evaluation.EvaluationResults(results=results)
    return results


@prefect.task
async def evaluate_dataset(dataset_name: str, bot_model_name: str):
    async def process_inputs(inputs: dict) -> dict:
        example = JiraServiceDeskBotEvaluationExample(**inputs)
        response = await ask_jira_bot(example=example, model_name=bot_model_name)
        return response

    client = langsmith.Client()
    await client.aevaluate(
        process_inputs,
        data=dataset_name,
        evaluators=[evaluate_example],
        metadata={"Bot model": bot_model_name},
    )



import inspect
import textwrap

import cachetools.func
import langchain_core.messages
import langgraph.prebuilt
import pydantic

from ai_app.ai_utils import PreModelHook, astream_response_messages
from ai_app.apps.jira_service_desk.data import (
    JiraRequestPrimaryKey,
    get_saved_request_type_primary_keys,
)
from ai_app.apps.jira_service_desk.tools import (
    build_url_for_jira_request_type,
    fetch_detailed_jira_request_type_guide,
    fetch_jira_request_type_guides_closest_to_user_query,
)
from ai_app.config import get_ai_postgres_engine
from ai_app.core import Response, State
from ai_app.frontend import get_generating_text_span
from ai_app.tools import (
    InfoSecUserReminder,
    build_structured_response_tool,
    extract_tool_artifacts_from_agent_response,
    get_structured_output_from_agent_response,
    report_disclosure_of_sensitive_information,
    report_feedback,
)


def get_system_prompt() -> str:
    system_prompt = textwrap.dedent(f"""
        ### Role
        You are a Jira service desk expert in a bank. Your goal is to help employees find the most 
        appropriate Jira request type for their problem.

        ### Language instruction
        - Your primary instruction, above all else, is to detect the language of the user's very 
        last message (Azerbaijani, English, or Russian) and ALWAYS respond in that exact language.
        - DO NOT deviate from this rule, even if other context suggests a different language might be helpful.

        ### Steps
        - Use the RAG retriever "{fetch_jira_request_type_guides_closest_to_user_query.name}" tool to find relevant 
        Jira request types.
        - If there is an appropriate request type among those fetched based on the user's problem and use case,
        use the "{fetch_detailed_jira_request_type_guide.name}" tool to get more information about it.
        - If based on the retrieved information the request type is highly relevant to the user's problem:
            - Provide the non-prefilled request type URL from the "{fetch_detailed_jira_request_type_guide.name}" tool output.
            - Briefly explain why the request type fits the user's situation and how the request type is usually handled.
            - Explain how the user may try to resolve the issue by themselves.
            - What information is needed to create the request.
        If the request type is not sufficiently relevant, ask clarifying questions to understand the user's problem 
        and rerun the "{fetch_jira_request_type_guides_closest_to_user_query.name}" tool based on the gathered 
        information.
        - After finding a relevant request type, collect from the user the information needed to create 
        a prefilled URL for the suggested request type  via the "{build_url_for_jira_request_type.name}" 
        tool and provide it to the user. Focus on required fields that can be passed as URL parameters.

        ### Notes
        - Proactively ask for user's feedback and use the "{report_feedback.name}" tool to report it.
        - Use the "{build_structured_response_tool(None).name}" tool to generate the message for the user.
    """)  # noqa: E501
    return system_prompt


def build_prompts() -> list[langchain_core.messages.BaseMessage]:
    system_prompt = get_system_prompt()
    prompts = [
        langchain_core.messages.SystemMessage(system_prompt),
    ]
    return prompts


class BotResponse(pydantic.BaseModel):
    last_user_message_language: str = pydantic.Field(
        description=inspect.cleandoc("""
            The language of the last user message, which the bot should also use for its own 
            response.
        """),
    )
    content: str = pydantic.Field(description="The message content that the user will see.")


@cachetools.func.ttl_cache(ttl=60)
def get_cached_saved_request_type_primary_keys() -> set[JiraRequestPrimaryKey]:
    return get_saved_request_type_primary_keys(get_ai_postgres_engine())


class Bot:
    async def astream(self, model, messages, state: State | None = None):
        agent = langgraph.prebuilt.create_react_agent(
            model,
            prompt=get_system_prompt(),
            pre_model_hook=PreModelHook(model=model),
            tools=[
                fetch_jira_request_type_guides_closest_to_user_query,
                fetch_detailed_jira_request_type_guide,
                build_url_for_jira_request_type,
                # report_disclosure_of_sensitive_information,
                report_feedback,
                build_structured_response_tool(BotResponse),
            ],
            response_format=BotResponse,
        )
        response = None
        async for response in astream_response_messages(
            agent, messages, stream_ai_tokens=True, with_spinner=False
        ):
            yield Response(
                content=response.formatted_response + get_generating_text_span(),
                messages=response.state["messages"],
            )

        if not response:
            return

        bot_response = get_structured_output_from_agent_response(response.state)
        yield Response(
            content=response.formatted_response + bot_response.content,
            messages=response.state["messages"],
            content_for_langsmith=bot_response.content,
        )

    async def respond(self, model, messages, state: State | None = None) -> Response:
        response = Response(content="")
        async for response in self.astream(model, messages, state):
            pass

        return response



import langgraph.prebuilt

from ai_app.ai_utils import PreModelHook
from ai_app.config import get_sql_toolkit
from ai_app.core import State


class Bot:
    async def respond(
        self,
        model,
        model_input,
        state: State,
        connection_context_name: str,
    ) -> str:
        toolkit = get_sql_toolkit(connection_context_name)
        agent = langgraph.prebuilt.create_react_agent(
            model=model,
            pre_model_hook=PreModelHook(model),
            tools=toolkit.get_tools(),
            prompt=toolkit.prompt,
        )
        response: dict = await agent.ainvoke({"messages": model_input})
        response_message = response["messages"][-1].content
        return response_message



import textwrap
from typing import Iterable, override

import langchain.chat_models
import langchain.hub
import langchain.tools
import langchain_community.agent_toolkits.sql.toolkit
import langchain_community.utilities
import sqlalchemy as sa

from .base import BaseToolkit


class SqlToolkit(BaseToolkit):
    def __init__(
        self,
        engine: sa.Engine,
        schema: str | None = None,
        include_tables: Iterable[str] | None = None,
        sql_checker_model: langchain.chat_models.base.BaseChatModel | None = None,
        top_k_query_rows: int = 20,
        prompt: str | None = None,
    ):
        self.database = langchain_community.utilities.SQLDatabase(
            engine=engine,
            schema=schema,
            include_tables=include_tables,
            sample_rows_in_table_info=top_k_query_rows,
            lazy_table_reflection=True,
        )
        if sql_checker_model:
            self.toolkit = langchain_community.agent_toolkits.sql.toolkit.SQLDatabaseToolkit(
                db=self.database,
                llm=sql_checker_model,
            )
        else:
            self.toolkit = None

        if not prompt:
            # prompt_template = langchain.hub.pull("langchain-ai/sql-agent-system-prompt")
            # prompt = prompt_template.format(dialect=self.database.dialect, top_k=top_k_query_rows)
            prompt = f"""
                You are an agent designed to interact with a SQL database.
                Given an input question, create a syntactically correct {self.database.dialect} query to run, then look 
                at the results of the query and return the answer.
                Unless the user specifies a specific number of examples they wish to obtain, always limit your query to 
                at most {top_k_query_rows} results.
                You can order the results by a relevant column to return the most interesting examples in the database.
                Never query for all the columns from a specific table, only ask for the relevant columns given the question.
                You have access to tools for interacting with the database.
                Only use the below tools. Only use the information returned by the below tools to construct your final answer.
                You MUST double check your query before executing it. If you get an error while executing a query, rewrite 
                the query and try again.

                DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.

                To start you should ALWAYS look at the tables in the database to see what you can query.
                Do NOT skip this step.
                Then you should query the schema of the most relevant tables.

                When referring to table columns, always quote them with double quotes, like "Metric" or "Date".
            """  # noqa: E501

        self.prompt = textwrap.dedent(prompt)

    @override
    def get_tools(
        self, sql_checker_model: langchain.chat_models.base.BaseChatModel | None = None
    ) -> list[langchain.tools.BaseTool]:
        if sql_checker_model:
            toolkit = langchain_community.agent_toolkits.sql.toolkit.SQLDatabaseToolkit(
                db=self.database,
                llm=sql_checker_model,
            )
        else:
            toolkit = self.toolkit

        if not toolkit:
            raise ValueError(
                "An SQL checker model should be provided either during toolkit initialization "
                "or get_tools call."
            )

        tools = toolkit.get_tools()
        return tools



from typing import override

import gradio as gr
import langchain_core.language_models
import langchain_core.messages

from ai_app.apps.jira_service_desk.backend import Bot
from ai_app.config import get_config
from ai_app.core import BaseApp, Response, State
from ai_app.frontend import build_model_choice_dropdown


class App(BaseApp):
    name = "Jira service desk"
    requires_auth = False

    def __init__(self):
        self.bot = Bot()

    async def astream(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
    ):
        messages = state.messages + [message]
        async for response in self.bot.astream(model, messages, state=state):
            yield response

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks() as app:
            model_choice = model_choice or build_model_choice_dropdown()
            self.build_gradio_chat_interface(
                model_choice=model_choice,
                description="The chatbot will help the user find appropriate Jira request type.",
                examples=(
                    None
                    if get_config().is_prod
                    else [
                        ["I want to change my password"],
                        ["I need access to a database"],
                        ["I received a phishing email"],
                        ["A system requires additional processing resources"],
                    ]
                ),
            )

        return app



from typing import override

import bir_mcp
import langchain.tools

from ai_app.tools.base import BaseToolkit


class JiraToolkit(BaseToolkit):
    def __init__(
        self,
        token: str,
        url: str,
        api_version: int = 2,
        max_tool_output_length: int | None = None,
        **kwargs,  # The kwargs from JiraParameters, need to use them for tenacity manually instead of passing to atlassian.Jira.
    ):
        super().__init__()
        self.mcp = bir_mcp.atlassian.Atlassian(
            token=token, url=url, api_version=api_version, **kwargs
        )
        self.max_tool_output_length = max_tool_output_length

    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = [
            self.mcp.get_jira_issue_overview,
        ]
        tools = [
            bir_mcp.utils.to_langchain_tool(tool, max_output_length=self.max_tool_output_length)
            for tool in tools
        ]
        return tools



import logging

import typer
import uvicorn

from ai_app.app import build_app
from ai_app.config import Stage, get_config


def launch_app(environment: Stage = Stage.dev, workers: int = 1):
    get_config().setup_logging()
    app = build_app()
    logging.info("Launching app server")
    uvicorn.run(
        app,
        host="0.0.0.0" if environment == Stage.prod else "localhost",
        port=get_config().port,
        workers=workers,
    )


def main():
    typer.run(launch_app)


if __name__ == "__main__":
    main()
